#include <fcntl.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <vca_fdpoll.h>
#include <vca_net.h>
#include <vcasdo_msg.h>

#include <ul_dbufflog.h>
#include <ul_log.h>
#include <ul_logreg.h>

#include "fcpickle_impl.h"
#include "candtgmsg.h"
#include "sdodownloadconfirmmsg.h"
#include "sdouploadconfirmmsg.h"
#include "sdodownloadrequestmsg.h"
#include "sdouploadrequestmsg.h"
#include "servicesetrawmsgparamsconfirm.h"
#include "servicesetrawmsgparamsrequest.h"

#define CAN_DTG_MSG_CLASS "ocera.rtcan.msg.CANDtgMsg"
#define SDO_UP_REQ_CLASS "ocera.rtcan.msg.SDOUploadRequestMsg"
#define SDO_DN_REQ_CLASS "ocera.rtcan.msg.SDODownloadRequestMsg"
#define SDO_UP_CNF_CLASS "ocera.rtcan.msg.SDOUploadConfirmMsg"
#define SDO_DN_CNF_CLASS "ocera.rtcan.msg.SDODownloadConfirmMsg"
#define SERVICE_SET_RAW_MSG_PARAMS_CONFIRM_CLASS "ocera.rtcan.msg.ServiceSetRawMsgParamsConfirm"
#define SERVICE_SET_RAW_MSG_PARAMS_REQUEST_CLASS "ocera.rtcan.msg.ServiceSetRawMsgParamsRequest"

#include "socknode.h"

//====================== logging support =======================

extern UL_LOG_CUST(ulogd_socknode);

//-------------------------------------------
static void fatal_error(const char *format, ...)
{
    va_list ap; 
    va_start(ap, format); 
    ul_vloglev(UL_LOGL_FATAL, format, ap);
    va_end(ap);
    clean_up();
    exit(EXIT_FAILURE);
}

//=============================================================

static void FdNodeSrvSock_ready(struct vcaFdNode_t *nd, int event_type, struct vcaFdPoll_t *p)
{
    FdNodeSrvSock_t *srv = UL_CONTAINEROF(nd, FdNodeSrvSock_t, fdnode);
    FdNodeCliSock_t *cli = FdNodeCliSock_init(NULL);
    ul_logdeb("request on server socket: %i\n", srv->fdnode.fd);
    if(cli) {
        int sin_size = sizeof(struct sockaddr_in);
        struct sockaddr_in cli_addr;
        cli->fdnode.fd = accept(srv->fdnode.fd, (struct sockaddr *)&cli_addr, &sin_size);
        if(cli->fdnode.fd < 0) {
            ul_logerr("FdNodeSrvSock_ready() accept() ERROR\n");
        }
        else {
            if(fcntl(cli->fdnode.fd, F_SETFL, O_NONBLOCK) < 0) {
                fatal_error("ERROR fcntl\n");
            }
            ul_logmsg("Client request from %s:%d\n", inet_ntoa(cli_addr.sin_addr), ntohs(cli_addr.sin_port));
            cli->cannet = srv->cannet;
            //char msg[] = "HELLO from canmond.";
            //write(cli->fdnode.fd, msg, sizeof(msg) - 1);
            // add client to fdpoll
            vcaFdPoll_insert(p, &(cli->fdnode));
        }
    }
}

FdNodeSrvSock_t* FdNodeSrvSock_init(FdNodeSrvSock_t *sock)
{
    if(sock == NULL) {
        sock = (FdNodeSrvSock_t*)malloc(sizeof(FdNodeSrvSock_t));
        if(!sock) fatal_error("memory allocation ERROR\n");
    }
    vcaFdNode_init(&(sock->fdnode), -1, VCAFD_RD);
    sock->fdnode.fd = -1;
    sock->fdnode.ready = FdNodeSrvSock_ready;
    sock->cannet = NULL;
    return sock;
}
 
void FdNodeSrvSock_destroy(FdNodeSrvSock_t *sock)
{
    //if(fdnode.fd > 0) close(fd);
    sock->fdnode.ready = NULL;
}

//------------------------------------------------------------

typedef struct socknodeSDORequest_t
{
  vcaSDORequest_t sdorq;
  ul_list_node_t sock_peers;
  FdNodeCliSock_t *cli;
} socknodeSDORequest_t;


UL_LIST_CUST_DEC(socknodeSDORequestQue, FdNodeCliSock_t, socknodeSDORequest_t,
                 rq_que, sock_peers)


void FdNodeCliSock_vcaSDORequest_finished(struct vcaSDORequest_t *sdorq);

socknodeSDORequest_t *socknodeSDORequest_init(
    socknodeSDORequest_t *req, FdNodeCliSock_t *cli,
    int is_upload, int index, int subindex, void *data, int data_len)
{
    if(!req){
        req = (socknodeSDORequest_t*)malloc(sizeof(socknodeSDORequest_t));
        if(!req){
            ul_logerr("memory allocation ERROR\n");
            return NULL;
        }
    }
    vcaSDORequest_init(&req->sdorq, is_upload, index, subindex, data, data_len);
    req->sdorq.finished_callback = FdNodeCliSock_vcaSDORequest_finished;
    req->cli = cli;
    if(cli) {
        socknodeSDORequestQue_insert(cli, req);
    }
    return req;
}

void socknodeSDORequest_destroy(socknodeSDORequest_t *req)
{
    if(req->sdorq.is_pending)
         vcaSDORequest_abort(&req->sdorq);

    if(req->cli){
        socknodeSDORequestQue_delete(req->cli, req);
        req->cli = NULL;
    }
    vcaSDORequest_destroy(&req->sdorq);
}


static int pickle_sdo_request(ul_dbuff_t *dbuff, const struct vcaSDORequest_t *sdorq)
{
    SDOConfirmMsg_t *cmsg; // base class
    union {
        SDOUploadConfirmMsg_t umsg;   
        SDODownloadConfirmMsg_t dmsg;
    } msg;
    
    if(sdorq->is_upload) {
        SDOUploadConfirmMsg_init(&msg.umsg);
        cmsg = &(msg.umsg.super);
    }
    else { 
        SDODownloadConfirmMsg_init(&msg.dmsg);
        cmsg = &(msg.dmsg.super);
    }
    
    //cmsg->super.srvcliCobId = fsm->srvcli_cob_id;
    //cmsg->super.clisrvCobId = fsm->clisrv_cob_id;
    cmsg->super.node = sdorq->node;
    cmsg->super.index = sdorq->index;
    cmsg->super.subindex = sdorq->subindex;
    if(sdorq->state == sdofsmAbort) {
        uint32_t code = 0;
        cmsg->code = sdorq->err_no;
        cmsg->type = SDOConfirmMsg_MSG_ABORT;
        cmsg->errmsg = vcasdo_abort_msg(code);
    }
    else if(sdorq->state == sdofsmError) {
        cmsg->type = SDOConfirmMsg_MSG_ERROR;
        cmsg->code = sdorq->err_no;
        cmsg->errmsg = vcasdo_abort_msg(cmsg->code);
    } 
    else { 
        cmsg->type = SDOConfirmMsg_MSG_OK;
    }

    if(sdorq->is_upload) {
        int ret;
        msg.umsg.data = sdorq->data.data;
        msg.umsg.data_arraylen = sdorq->data.len;
        ret = SDOUploadConfirmMsg_getPacketBufferSize(&msg.umsg);
        if(ret < 0) {
            ul_logerr("sdo2msg() SDOUploadConfirmMsg_getPacketBufferSize() ERROR(%i)\n", ret);
            ul_dbuff_set_len(dbuff, 0);
        }
        ul_dbuff_set_len(dbuff, ret);
        ret = SDOUploadConfirmMsg_toNet(&msg.umsg, dbuff->data, dbuff->len);
        if(ret < 0) {
            ul_logerr("sdo2msg() SDOUploadConfirmMsg_toNet() ERROR(%i)\n", ret);
            ul_dbuff_set_len(dbuff, 0);
        }
    }
    else {
        int ret;
        ul_dbuff_set_len(dbuff, SDODownloadConfirmMsg_getPacketBufferSize(&msg.dmsg));
        ret = SDODownloadConfirmMsg_toNet(&msg.dmsg, dbuff->data, dbuff->len);
        if(ret < 0) {
            ul_logerr("sdo2msg() SDODownloadConfirmMsg_toNet() ERROR(%i)\n", ret);
            ul_dbuff_set_len(dbuff, 0);
        }
    }

    return dbuff->len > 0;
}

void FdNodeCliSock_vcaSDORequest_finished(struct vcaSDORequest_t *sdorq)
{
    socknodeSDORequest_t *sockrq = UL_CONTAINEROF(sdorq, socknodeSDORequest_t, sdorq);
    FdNodeCliSock_t *cli = sockrq->cli;
    ul_logdeb("FdNodeCliSock_vcaSDORequest_finished called\n");

    // FIXME: The SDO transfer result has to be pickled and sent to the monitor
    // pickle SDO a request response
    {
        ul_dbuff_t db;
        ul_dbuff_init(&db, 0);
        if(pickle_sdo_request(&db, sdorq)) {
            write(cli->fdnode.fd, db.data, db.len);
        }
        ul_dbuff_destroy(&db);
    }
    socknodeSDORequest_destroy(sockrq);
    free(sockrq);
}

void FdNodeCliSock_msgrec_fnc(FdNodeCliSock_t *sock, int event_code, canmsg_t *canmsg)
{
    int len;
    CANDtgMsg_t msg;
    CANDtgMsg_init(&msg);
    len = CANDtgMsg_getPacketBufferSize(NULL); // fixed len message
    if(len > 0) { // this should be allways true
        ul_dbuff_t dbuff;
        ul_dbuff_init(&dbuff, 0);
        ul_dbuff_set_len(&dbuff, len);
        msg.flags = canmsg->flags;
        msg.id = canmsg->id;
        msg.length = canmsg->length;
        memcpy(&msg.data, canmsg->data, CAN_MSG_LENGTH);
        len = CANDtgMsg_toNet(&msg, dbuff.data, dbuff.len);
        if(len > 0) { // this should be allways true
            ul_logdeb("Sending RAW message to client ID: %x\n", msg.id);
            write(sock->fdnode.fd, dbuff.data, dbuff.len);
        }
        ul_dbuff_destroy(&dbuff);
    }
}
//------------------------------------------------------------


static void FdNodeCliSock_ready(struct vcaFdNode_t *nd, int event_type, struct vcaFdPoll_t *p)
{
    FdNodeCliSock_t *cli = UL_CONTAINEROF(nd, FdNodeCliSock_t, fdnode);
    int fd = cli->fdnode.fd;
    int nread;
    int pickle_result = 0;
    ul_dbuff_t readdb;

    if(fd < 0) return;
    
    ul_logdeb("data on client socket: %i\n", fd);

    ioctl(fd, FIONREAD, &nread);
    if(nread == 0) {
        // client closed socket
        vcaFdPoll_detach(p, &(cli->fdnode));
	FdNodeCliSock_destroy(cli);
        free(cli);
        return;
    }

    ul_dbuff_init(&readdb, 0);

    if(ul_dbuff_set_len(&readdb, nread) == nread) {
        nread = read(fd, readdb.data, nread);
    }
    else {
        fatal_error("Cann't allocate memory for client data\n");
    }
    
    {
        ul_dbuff_t *msgdb = &(cli->msgdb);
        int i;
        // copy data from readdb to msgdb and try to find object packet start
        for(i=0; i<readdb.len; i++) {
            unsigned char b = readdb.data[i];
            if(msgdb->len < FCPICKLE_PACKET_HEAD_SIZE) {
                // read packet head
                ul_dbuff_append_byte(msgdb, b);
                if(msgdb->len == FCPICKLE_PACKET_HEAD_SIZE) {
                    //ul_logdeb("got packet head:\n");
                    //ul_dbuff_log_hex(msgdb, LOG_TRASH);
                    // may be the last byte of HEAD came
                    fcpickle_packet_head_t head;
                    // check if it could be a object packet head
                    if(fcpickle_getHead(&head, msgdb->data, msgdb->len) < 0) {
                        // NO it is not
                        // shift data in msgdb one byte right
                        int j;
                        for(j=0; j<FCPICKLE_PACKET_HEAD_SIZE-1; j++) msgdb->data[j] = msgdb->data[j+1];
                        msgdb->len--;
                    }
                    else {
                        // I've got the objec packet head 
                        cli->packet_len = head.packetSize;
                        ul_logdeb("expected packet size: %i\n", head.packetSize);
                    }
                }
            }
            else if(msgdb->len < cli->packet_len) {
                // read rest of packet
                ul_dbuff_append_byte(msgdb, b);
            }
            if(msgdb->len == cli->packet_len) {
                // now I have whole packet
                const char *class_name = fcpickle_getClassName(msgdb->data, msgdb->len);
                vcaNet_t *net = cli->cannet;
                ul_logdeb("Object from a client socket received: %s\n", class_name);
                if(strcmp(class_name, CAN_DTG_MSG_CLASS) == 0) {
                    CANDtgMsg_t candtg;
                    CANDtgMsg_init(&candtg);
                    pickle_result = CANDtgMsg_fromNet(&candtg, msgdb->data, msgdb->len);
                    if(pickle_result > 0) {
                        canmsg_t canmsg;
                        memset(&canmsg, 0, sizeof(canmsg));
                        canmsg.flags = candtg.flags; 
                        canmsg.id = candtg.id;
                        canmsg.length = candtg.length;
                        memcpy(&canmsg.data, &candtg.data, canmsg.length);
                        vcaNet_send_can_msg(net, &canmsg);
                    }
                }
                else if(strcmp(class_name, SDO_UP_REQ_CLASS) == 0 || strcmp(class_name, SDO_DN_REQ_CLASS) == 0) {
                    // all structures have a same base fields
                    int is_upload = 0;
                    SDOMsgBase_t *msg = NULL;
                    SDOUploadRequestMsg_t umsg;
                    SDODownloadRequestMsg_t dmsg;
                
                    if(strcmp(class_name, SDO_UP_REQ_CLASS) == 0) {
                        ul_logdeb("%s() Upload request '%s' unpickling packet.\n", __func__, class_name);
                        pickle_result = SDOUploadRequestMsg_fromNet(&umsg, msgdb->data, msgdb->len);
                        msg = &(umsg.super);
                        is_upload = 1;
                    }
                    else if(strcmp(class_name, SDO_DN_REQ_CLASS) == 0) {
                        ul_logdeb("%s() Download request '%s' unpickling packet.\n", __func__, class_name);
                        pickle_result = SDODownloadRequestMsg_fromNet(&dmsg, msgdb->data, msgdb->len);
                        msg = &(dmsg.super);
                        is_upload = 0;
                    }
                    else {
                        ul_loginf("%s() unknown SDO message object '%s', no FSM created.\n", __func__, class_name);
                        ul_dbuff_destroy(&readdb);
                        return;
                    }
                    
                    ul_logdeb("%s() '%s' successfully unpickled.\n", __func__, class_name);
                    {
                        // create SDO request and add it to node SDO requests list
                        socknodeSDORequest_t *req = NULL;
                        int node = msg->node;
                        vcaCANNode_t *nd = vcaNet_find_cannode(net, node, 1 /* create */);

                        if(nd == NULL) fatal_error("%s() vcaNet_find_cannode(..., %i, ...) ERROR\n", __func__, node);

                        if(is_upload) {
                            req = socknodeSDORequest_init(NULL, cli, is_upload, msg->index, msg->subindex,
                                                     NULL, 0);
                        } 
                        else {
                            req = socknodeSDORequest_init(NULL, cli, is_upload, msg->index, msg->subindex,
                                                     dmsg.data, dmsg.data_arraylen); 
                        }
                        if(req) {
                            vcaCANNode_addSDORequest(nd, &req->sdorq);
                        }
                    }
                }
                else if(strcmp(class_name, SERVICE_SET_RAW_MSG_PARAMS_REQUEST_CLASS) == 0) {
                    ServiceSetRawMsgParamsRequest_t req;
                    ul_logdeb("%s() request '%s' unpickling packet.\n", __func__, class_name);
                    ServiceSetRawMsgParamsRequest_init(&req);
                    pickle_result = ServiceSetRawMsgParamsRequest_fromNet(&req, msgdb->data, msgdb->len);
                    if(pickle_result > 0) {
                        int ok = 0;
                        int len;
                        ServiceSetRawMsgParamsConfirm_t msg;

                        ul_logdeb("\tcommand: %i\n", req.command);
                        ul_logdeb("\tid: %x\n", req.id);
                        ul_logdeb("\tmask: %x\n", req.mask);
                        if(req.command == ServiceSetRawMsgParamsRequest_ADD) {
			    if(vcaNet_msg_rec_connect(net, &cli->rough_canmsg_hub, req.id, req.mask)>=0)
			        ok=1;
			    else
			        ok=0;
                        }
                        else if(req.command == ServiceSetRawMsgParamsRequest_REMOVE_ALL) {
			    evc_rx_hub_disconnect_all(&cli->rough_canmsg_hub);
			    ok=1;
                        }
                        ServiceSetRawMsgParamsConfirm_init(&msg);
                        msg.super.code = (ok)? ServiceConfirmMsg_OK: ServiceConfirmMsg_ERROR;
                        msg.super.errmsg = "not implemented yet";
                        len = ServiceSetRawMsgParamsConfirm_getPacketBufferSize(&msg);
                        if(len > 0) { // this should be allways true
                            ul_dbuff_t dbuff;
                            ul_dbuff_init(&dbuff, 0);
                            ul_dbuff_set_len(&dbuff, len);
                            len = ServiceSetRawMsgParamsConfirm_toNet(&msg, dbuff.data, dbuff.len);
                            if(len > 0) { // this should be allways true
                                write(fd, dbuff.data, dbuff.len);
                            }
                            ul_dbuff_destroy(&dbuff);
                        }
                    }
                }
                if(pickle_result < 0) {
                    ul_logerr("%s() unpickling object ERROR %i\n", __func__, -pickle_result);
                    ul_dbuff_destroy(&readdb);
                    return;
                }
                
                ul_dbuff_set_len(msgdb, 0);
                cli->packet_len = 0;
            }
        }
        if(cli->packet_len != 0) {
            ul_logdeb("msgdb:\n");
            ul_dbuff_log_hex(msgdb, UL_LOGL_TRASH);
        }
    }
    ul_dbuff_destroy(&readdb);
}

FdNodeCliSock_t* FdNodeCliSock_init(FdNodeCliSock_t *sock)
{
    if(sock == NULL) {
        sock = (FdNodeCliSock_t*)malloc(sizeof(FdNodeCliSock_t));
        if(!sock) fatal_error("memory allocation ERROR\n");
    }
    vcaFdNode_init(&(sock->fdnode), -1, VCAFD_RD);
    sock->packet_len = 0;
    ul_dbuff_init(&(sock->msgdb), 0);
    sock->fdnode.ready = FdNodeCliSock_ready;

    evc_rx_hub_init(&sock->rough_canmsg_hub,(evc_rx_fnc_t*)FdNodeCliSock_msgrec_fnc,sock);
    
    socknodeSDORequestQue_init_head(sock);
    return sock;
}

void FdNodeCliSock_destroy(FdNodeCliSock_t *sock)
{
    socknodeSDORequest_t *req;
    
    ul_list_for_each(socknodeSDORequestQue,sock,req){
        socknodeSDORequest_destroy(req);
        free(req);
    }

    evc_rx_hub_done(&sock->rough_canmsg_hub);

    ul_dbuff_destroy(&(sock->msgdb));
    sock->fdnode.ready = NULL;
}


