#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <stdarg.h>
#include <string.h>

#include <can_vca.h>
#include <eds_utils.h>
#include <vcasdo_msg.h>
#include <ul_log.h>

#include "vca_chsigcon.h"

//======================= logging support =========================

extern UL_LOG_CUST(ulogd_vca);

//=================================================================

//-----------------------------------------------------------------
static int line_no = 0; // counter of lines read from EDS
static int get_line(FILE *in, ul_dbuff_t *line)
{
    int c;
    ul_dbuff_set_len(line, 0);
    while(1) {
        c = getc(in);
        if(c == EOF) break;
        ul_dbuff_append_byte(line, (unsigned char)c);
        if(c == '\n') break;
    }
    if(line->len) line_no++;
    ul_dbuff_append_byte(line, '\0');
    return line->len - 1;   // ignore terminating zero
}
//-----------------------------------------------------------------
// reads object definition from EDS file
// object definition lines after [....]
// after return line contains next unprocessed line from EDS file
static void get_od_item_def(FILE *in, ul_dbuff_t *line, vcaod_object_t *item)
{
    int l=0;
    int is_subindex;
    int type_size = 0;
    int default_val_is_set = 0;

    if(!item) return;
    is_subindex = (item->subcnt == 0);
    while(1) {
        if(!line->len || !line->data[0]) {
            l = get_line(in, line); // get new line only if the old is processed
            if(!l) break; // EOF
        }
        ul_dbuff_trim(line);
        if(line->data[0] == '[') break; // next index/subindex definition
        l = ul_str_cpos(line->data, '=', 0);
        if(l >= 0) {
            line->data[l] = '\0'; 
            if(vca_strmatch(cpuc2pc(line->data), "ParameterName") > 0) {
                ul_str_ncpy(pc2puc(item->name), line->data + l + 1, VCAOD_OBJECT_NAME_LEN);
            }
            else if(vca_strmatch(cpuc2pc(line->data), "SubNumber") > 0) { 
                errno = 0;
                item->subcnt = strtol(cpuc2pc(line->data + l + 1), NULL, 0);
                if(errno) item->subcnt = 0;
            }
            else if(vca_strmatch(cpuc2pc(line->data), "ObjectType") > 0) { 
                errno = 0;
                item->object_type = strtol(cpuc2pc(line->data + l + 1), NULL, 0);
                if(errno) item->object_type = 0;
                if(is_subindex) switch(item->object_type) {
                    case 0:
                    case vcaod_object_type_VAR: break;
                    default: 
                        ul_logerr("[%04x:%02x] subindex type should not be 0x%x (%s)\n", 
                                item->index, item->subindex, item->object_type, od_item_type2str(item->object_type));
                        ul_logerr("\tchanging it to 0x7 (VAR)\n"); 
                        item->object_type = vcaod_object_type_VAR;
                        break;
                }
                type_size = vcaod_get_object_data_size(item, 0);
            }
            else if(vca_strmatch(cpuc2pc(line->data), "DataType") > 0) { 
                errno = 0;
                item->data_type = strtol(cpuc2pc(line->data + l + 1), NULL, 0);
                if(errno) item->data_type = 0;
            }
            else if(vca_strmatch(cpuc2pc(line->data), "AccessType") > 0) {
                unsigned char *pc = line->data + l + 1;
                if(vca_strmatch(cpuc2pc(pc), "ro") >= 0) item->access = vcaod_access_RO;
                else if(vca_strmatch(cpuc2pc(pc), "wo") >= 0) item->access = vcaod_access_WO;
                else if(vca_strmatch(cpuc2pc(pc), "rw") >= 0) item->access = vcaod_access_RW;
                else if(vca_strmatch(cpuc2pc(pc), "rwr") >= 0) item->access = vcaod_access_RWR;
                else if(vca_strmatch(cpuc2pc(pc), "rww") >= 0) item->access = vcaod_access_RWW;
                else if(vca_strmatch(cpuc2pc(pc), "const") >= 0) item->access = vcaod_access_CONST;
            }
            else if(vca_strmatch(cpuc2pc(line->data), "DefaultValue") > 0) {
                od_item_set_value_as_str(item, cpuc2pc(line->data + l + 1));
                default_val_is_set = 1;
            }
            else if(vca_strmatch(cpuc2pc(line->data), "PDOMapping") > 0) {
                long li;
                errno = 0;
                li = strtol(cpuc2pc(line->data + l + 1), NULL, 0);
                item->flags &= ~VCAOD_OBJECT_FLAG_PDO_MAPPING;
                if(!errno) if(li) item->flags |= VCAOD_OBJECT_FLAG_PDO_MAPPING;
            }
            else {
                #ifdef _DEBUG_FULL
                ul_logdeb("[%04x:%02x] get_od_item_def() - line %i: flushing: %s\n", item->index, item->subindex, line_no, line->data);
                #endif
            }
            //LowLimit=0x01, HighLimit=0x088, is not implemented yet
        }
        line->data[0] = 0;
    }
    if(!default_val_is_set) {
        od_item_set_value_as_str(item, "");
    }
}
//-----------------------------------------------------------------
/**
 * vcaod_load_eds - opens file and create new OD acording to its contens
 * @odroot:    root, which will contain loaded EDS
 * @file_name: name of file to load
 * Returns:    zero in case of success
 * Header:     eds_utils.h
 */
int vcaod_load_eds(vcaod_root_t *odroot, const char* file_name)
{
    FILE *in;
    ul_dbuff_t line;
    line_no = 0;
    ul_loginf("Loading EDS from file '%s'\n", file_name);
    if(!vcaod_is_empty(odroot)) {
        // free OD
        vcaod_od_free(odroot); // may be this should be better
        //ul_logerr("vcaod_load_eds: OD root is not empty");
        //return -1;
    }
    in = fopen(file_name, "rb");
    if(!in) {
        ul_logmsg("vcaod_load_eds: can't open file %s for reading.\n", file_name);
        return -2;
    }
    
    ul_dbuff_init(&line, 0);
    while(1) {
        //int l=0;
        vcaod_object_t *item;
	
        if(!line.len || !line.data[0]) { 
            int l = get_line(in, &line); // get new line only if the old is processed
            if(!l) break; // eof
        }
        
        ul_dbuff_trim(&line);
        if(!line.data[0]) continue; // skip empty lines
        
        if(line.data[0] == '[') {
            // index
            unsigned ix;
            char *pc;
            ix = strtol(cpuc2pc(line.data+1), &pc, 16);
            if(*pc != ']') { // other than index declaration
                if(ul_str_pos(pc2puc(pc), cpc2puc("sub"), 0) >= 0) {
                    ul_logerr("EDS read error at line %i: '%s'\n"
                                   "subindex declaration should not be here\n", 
                                   line_no, line.data);
                }
                else {
                    #ifdef _DEBUG_FULL
                    ul_logdeb("line %i: flushing: %s\n", line_no, line.data);
                    #endif
                }
                line.data[0] = 0; // mark line as processed
                continue;
            }
            
            *pc = '\0'; // cutoff trailing ']'
            
            // index declaration
            #ifdef _DEBUG_FULL
            ul_logdeb("+++++ processing line %i: %s\nparsing index [%04x]\n", line_no, line.data, ix);
            #endif
            item = od_item_init(NULL);
            item->index = ix;
            line.data[0] = 0;
            get_od_item_def(in, &line, item);
            if(item->subcnt > 0) {
                // read subindicies
                int subcnt = item->subcnt;
                // make space also for optional FF subindex
                item->subobjects = malloc(sizeof(vcaod_object_t) * subcnt);
                if(item->subobjects) {
                    int i;
                    for(i=0; i<subcnt; i++) od_item_init(item->subobjects + i);
                    while(line.len) {
                        ul_dbuff_trim(&line);
                        if(line.data[0] == '[' && ul_str_pos(line.data, cpc2puc("sub"), 0) == 5) {
                            int subix = strtol(cpuc2pc(line.data+8), NULL, 16);
                            #ifdef _DEBUG_FULL
                            ul_logdeb("----- processing line %i: %s\nparsing index [%04x:%02x]\n", line_no, line.data, item->index, subix);
                            #endif
                            if(subix < subcnt || subix == 0xFF) {
                                int six = subix;
                                if(subix == 0xFF) {
                                    six = subcnt; // optional subindex FF (see CANopen docs)
                                    if(subcnt == item->subcnt) {
                                        // append only one FF subindex
                                        item->subcnt++;
                                        item->subobjects = realloc(item->subobjects, sizeof(vcaod_object_t) * item->subcnt);
                                        od_item_init(item->subobjects + subcnt);
                                    }
                                }
                                item->subobjects[six].index = item->index;
                                item->subobjects[six].subindex = subix;
                                line.data[0] = 0;
                                get_od_item_def(in, &line, item->subobjects + six);
                            }
                            else {
                                ul_logerr("EDS parser error line %i: %s\nsubindex [%02x] out of range (%x)\n", line_no, line.data, subix, subcnt);
                            }
                        }
                        else {
                            // next object
                            break;
                        }
                    }
                    // check if all subindicies are filled
                    for(i=0; i<subcnt; i++) if(item->subobjects[i].subindex < 0) {
                        ul_logerr("EDS read error at line %i: index[%04x]subobject[%02x] not filled\n", line_no, ix, i);
                    }
                }
            }
            vcaod_insert(odroot, item);
        }        
        else {
            line.data[0] = 0; // mark line as processed
        }
    }
    ul_dbuff_destroy(&line);
    return 0;
}
//-----------------------------------------------------------------
/**
 * vcaod_load_hds - opens HDS file and connect dinfo to appropriate objects in OD
 * @odroot:    root, which will contain OD
 * @file_name: name of file to load
 * Returns:    zero in case of success
 * Header:     eds_utils.h
 */
int vcaod_load_hds(vcaod_root_t *odroot, sui_dtree_memdir_t *dinfotreeroot, 
                   const char* file_name, const char* dll_search_path)
{ 
    FILE *in;
    ul_dbuff_t line;
    ul_dbuff_t token;
    line_no = 0;
    ul_loginf("Loading HDS from file '%s'\n", file_name);
    sui_dtreemem_dll_search_path = dll_search_path;
    in = fopen(file_name, "rb");
    if(!in) {
        ul_logerr("%s: can't open file %s for reading.\n", __PRETTY_FUNCTION__, file_name);
        return -2;
    }
    
    ul_dbuff_init(&line, 0);
    ul_dbuff_init(&token, 0);
    while(1) {
        int l = get_line(in, &line); // get new line only if the old is processed
        if(!l) break; // eof
        
        ul_dbuff_trim(&line);
        if(line.len == 0) continue; // skip empty lines
        if(line.data[0] == '#') continue; // skip commented lines
        
        //ul_logdeb("parsing line: <begin>%s<end>\n", line.data);
        ul_dbuff_cut_token(&line, &token);
        ul_dbuff_append_byte(&token, '\0');
        ul_dbuff_trim(&line); // line is allready zero terminated
        // token should be something like 1AB4:0A
        if(token.len > 0) {
            unsigned int ix, subix;
            int pos = vca_gethex(cpuc2pc(token.data), &ix);
            //ul_logdeb("parsing token: '%s'\n", token.data);
            vca_gethex(cpuc2pc(token.data + pos + 1), &subix);
            ul_logdeb("Loading HW dinfo [%04x:%02x]->'%s'\n", ix, subix, line.data);
            {
                sui_dinfo_t *di;
                uint32_t err_code;
                vcaod_object_t *o = vcaod_find_object(odroot, ix, subix, &err_code);
                if(!o) {
                    ul_logerr("%s: cann't find object %04x:%02x in OD\n", __PRETTY_FUNCTION__, ix, subix);
                    ul_logerr("\t%s\n", vcasdo_abort_msg(err_code));
                    continue;
                }
                if(o->dinfo) {
                    ul_logerr("%s: object %04x:%02x already has dinfo asigned\n", __PRETTY_FUNCTION__, ix, subix);
                    continue;
                }
                if(sui_dtree_lookup(&(dinfotreeroot->dir), cpuc2pc(line.data), NULL, &di) != SUI_DTREE_FOUND) {
                    ul_logerr("HW dinfo [%04x:%02x]->'%s' NOT FOUND\n", ix, subix, line.data);
                    continue;
                }
                sui_dinfo_inc_refcnt(di);
                o->dinfo = di;
                o->flags &= ~VCAOD_OBJECT_FLAG_WEAK_DINFO;        
                ul_logdeb("%s: added HW dinfo (%p) to object %04x:%02x\n", __PRETTY_FUNCTION__, o->dinfo, ix, subix);
            }
        }
    }
    return 0;
}
//-----------------------------------------------------------------

