#include <ul_utmalloc.h>
#include <ul_htimer.h>
#include <ul_gsa.h>
#include <ul_gsacust.h>
#include <ul_log.h>
#include <stdlib.h>
#include <poll.h>

#include "vca_fdpoll.h"

//======================== logging support ==========================

extern UL_LOG_CUST(ulogd_vcafdpoll);

static void fatal_error(const char *format, ...)
{
    va_list ap;
    va_start(ap, format); 
    ul_vloglev(UL_LOGL_FATAL, format, ap);
    va_end(ap);
  #ifndef __RTL__
    exit(EXIT_FAILURE);
  #else /*__RTL__*/
    while (1) {
        /*Block all other activities*/
	pthread_wait_np ();
    }
  #endif /*__RTL__*/
}
//---------------------------------------------------------------------------

GSA_CUST_DEC(vcafd, vcaFdPoll_t, vcaFdNode_t, vca_fd_t,
	fdnode_list, fd, vcafd_cmp_fnc)

static inline int
vcafd_cmp_fnc(const vca_fd_t *a, const vca_fd_t *b)
{
  if (*a>*b) return 1;
  if (*a<*b) return -1;
  return 0;
}

GSA_CUST_IMP(vcafd, vcaFdPoll_t, vcaFdNode_t, vca_fd_t,
	fdnode_list, fd, vcafd_cmp_fnc, 0)

vcaFdNode_t* vcaFdNode_init(vcaFdNode_t *nd, int fd, int event_type)
{
    if(nd == NULL) {
        nd = (vcaFdNode_t*)malloc(sizeof(vcaFdNode_t));
        if(!nd) fatal_error("vcaFdNode_init() - memory allocation ERROR\n");
    }
    nd->fd = fd;
    nd->ready = NULL;
    nd->event_type = event_type;
    return nd;
}

void vcaFdNode_destroy(vcaFdNode_t *nd)
{
    /* FIXME: the actual code has to be inserted there */
}

vcaFdPoll_t* vcaFdPoll_init(struct vcaFdPoll_t *p)
{
    if(p == NULL) {
        p = (vcaFdPoll_t*)malloc(sizeof(vcaFdPoll_t));
        if(!p) fatal_error("vcaFdPoll_init() - memory allocation ERROR.\n");
    }
    vcafd_init_array_field(p);
    p->pollfd_array = NULL;
    p->pollfd_size = 0;
    return p;
}

void vcaFdPoll_destroy(vcaFdPoll_t *p)
{
    p->pollfd_size = 0;
    if(p->pollfd_array != NULL) {
        free(p->pollfd_array);
        p->pollfd_array = NULL;
    }
    if(p->fdnode_list.items != NULL) {
        free(p->fdnode_list.items);		/* FIXME: */
	p->fdnode_list.items = NULL;
    }
}

static void sync_pollfd_array(vcaFdPoll_t *p)
{
    vcafd_it_t it;
    struct pollfd *f;
    p->need_resync = 0;
    p->pollfd_size = p->fdnode_list.count;
    if(p->pollfd_size == 0) {
        if(p->pollfd_array) free(p->pollfd_array);
        p->pollfd_array = NULL;
        return;
    }
    if(p->pollfd_array) {
        p->pollfd_array = (struct pollfd*)realloc(p->pollfd_array, p->pollfd_size * sizeof(struct pollfd));
    }
    else {
        p->pollfd_array = (struct pollfd*)malloc(p->pollfd_size * sizeof(struct pollfd));
    }
    f = p->pollfd_array;
    ul_for_each_it(vcafd, p, it) {
        vcaFdNode_t *n = vcafd_it2item(&it);
        f->fd = n->fd;
        f->revents = 0;
        f->events = 0;
        if(n->event_type & VCAFD_RD) f->events |= POLLIN;
        if(n->event_type & VCAFD_WR) f->events |= POLLOUT;
        f++;
    }
}
 
void vcaFdPoll_insert(vcaFdPoll_t *p, vcaFdNode_t *nd)
{
    ul_logdeb("adding fd %i\n", nd->fd);
    if(nd->fd < 0) {
        ul_logerr("invalid fd: %i\n", nd->fd);
        return;
    }
    vcafd_insert(p, nd);
    p->need_resync = 1;
}

void vcaFdPoll_detach(vcaFdPoll_t *p, vcaFdNode_t *nd)
{
    ul_logdeb("detaching fd %i\n", nd->fd);
    vcafd_delete(p, nd);
    p->need_resync = 1;
}

int vcaFdPoll_poll(vcaFdPoll_t *p, ul_htim_diff_t timeout)
{
    int ret;
    int i;
    if(p->need_resync) {
        sync_pollfd_array(p);
    }
    ret = poll(p->pollfd_array, p->pollfd_size, timeout);
    if(ret > 0) {
        /* An event on one of the fds has occurred. */
        for(i=0; i<p->pollfd_size; i++) {
            int revents = p->pollfd_array[i].revents;
            int ev_type = 0;
            vcaFdNode_t *n;
            if(!revents) continue;
            if(revents & (POLLOUT)) ev_type |= VCAFD_WR;
            if(revents & (POLLIN)) ev_type |= VCAFD_RD;
            n = vcafd_at(p, i);
            if(n->ready) n->ready(n, ev_type, p);
            if(p->need_resync) break;
        }
    }
    return ret;
}

