/* tscan.c
 * Linux CAN-bus device driver.
 * Written by Arnaud Westenberg email:arnaud@wanadoo.nl
 * Rewritten for new CAN queues by Pavel Pisa - OCERA team member
 * email:pisa@cmp.felk.cvut.cz
 * This software is released under the GPL-License.
 * Version lincan-0.3  17 Jun 2004
 */

#include "../include/can.h"
#include "../include/can_sysdep.h"
#include "../include/main.h"
#include "../include/pcm3680.h"
#include "../include/sja1000p.h"

#include <linux/module.h>
#include "../include/modparms.h"
#include "../include/devcommon.h"

// for PVT, 25MHz, IRQ 33
/*
#define TSCAN_ADDR_IO   0x600ff180|0x14
#define TSCAN_DATA_IO   0x600ff180|0x16
#define IO_RANGE        0x4
#define TSCAN_CLK       25000
#define TSCAN_IRQ       33
*/

// for KorralKool, it is 25MHz, IRQ 32
/*
#define TSCAN_ADDR_IO   0x600FF000|0xE0
#define TSCAN_DATA_IO   0x600FF000|0xE2
#define IO_RANGE        0x4
#define TSCAN_CLK	25000
#define TSCAN_IRQ       32
*/

// for PROSOFT, it is 16MHz, IRQ 33
#define TSCAN_ADDR_IO   0x600FF000|0xE0
#define TSCAN_DATA_IO   0x600FF000|0xE2
#define IO_RANGE        0x4
#define TSCAN_CLK     16000
#define TSCAN_IRQ       33	

int tscan_request_io(struct candevice_t *candev);
int tscan_release_io(struct candevice_t *candev);
int tscan_reset(struct candevice_t *candev);
int tscan_init_hw_data(struct candevice_t *candev);
int tscan_init_chip_data(struct candevice_t *candev, int chipnr);
int tscan_init_obj_data(struct canchip_t *chip, int objnr);
void tscan_write_register(unsigned data, can_ioptr_t address);
unsigned tscan_read_register(can_ioptr_t address);
int tscan_program_irq(struct candevice_t *candev);

unsigned char *tscan_addr_io;
unsigned char *tscan_data_io;
static CAN_DEFINE_SPINLOCK(tscan_win_lock);

int tscan_request_io(struct candevice_t *candev)
{
	if (!can_request_io_region(TSCAN_ADDR_IO,IO_RANGE,DEVICE_NAME)) {
                CANMSG("Unable to open port: 0x%lx\n",candev->io_addr);
                return -ENODEV;
        }else {
                DEBUGMSG("Registered IO-memory: 0x%lx - 0x%lx\n", candev->io_addr, candev->io_addr + IO_RANGE - 1);
        }

	return 0;
}

int tscan_release_io(struct candevice_t *candev)
{
	can_release_io_region(TSCAN_ADDR_IO,IO_RANGE);
	iounmap(tscan_addr_io);
	iounmap(tscan_data_io);
	return 0;
}

int tscan_reset(struct candevice_t *candev)
{
	unsigned short i=0, chipnr;
	struct canchip_t *chip;

	DEBUGMSG("Resetting %s hardware ...\n", candev->hwname);

	for(chipnr=0;chipnr<candev->nr_sja1000_chips;chipnr++) {
		chip=candev->chip[chipnr];
		can_write_reg(chip, sjaMOD_RM, SJAMOD);
		udelay(1000);
		can_write_reg(chip, 0x00, SJAIER);
		udelay(1000);
		i=20;
		while (can_read_reg(chip, SJAMOD)&sjaMOD_RM){
			if(!i--) return -ENODEV;
			udelay(1000);
			can_write_reg(chip, 0, SJAMOD);
		}
		udelay(1000);
		can_write_reg(chip, sjaCDR_PELICAN, SJACDR);
		can_write_reg(chip, 0x00, SJAIER);
	}

	return 0;
}

#define RESET_ADDR 0x100
#define NR_82527 0
#define NR_SJA1000 1

int tscan_init_hw_data(struct candevice_t *candev)
{
	tscan_addr_io = (unsigned char *)ioremap(TSCAN_ADDR_IO,sizeof(unsigned char));
        tscan_data_io = (unsigned char *)ioremap(TSCAN_DATA_IO,sizeof(unsigned char));

	candev->io_addr=0x0;
	candev->res_addr=RESET_ADDR;
	candev->nr_82527_chips=NR_82527;
	candev->nr_sja1000_chips=NR_SJA1000;
	candev->nr_all_chips=NR_82527+NR_SJA1000;
	candev->flags &= ~CANDEV_PROGRAMMABLE_IRQ;

	DEBUGMSG("Memory region at 0x%lx assigned to sja1000 of driver %d/%s\n",
		candev->io_addr, candev->candev_idx, candev->hwname);

	return 0;
}

int tscan_init_chip_data(struct candevice_t *candev, int chipnr)
{
	sja1000p_fill_chipspecops(candev->chip[chipnr]);

	candev->chip[chipnr]->int_clk_reg = 0x0;
	candev->chip[chipnr]->int_bus_reg = 0x0;
	candev->chip[chipnr]->sja_cdr_reg = sjaCDR_CBP | sjaCDR_CLK_OFF;
	candev->chip[chipnr]->sja_ocr_reg = sjaOCR_MODE_NORMAL | sjaOCR_TX0_LH;

	candev->chip[chipnr]->clock = TSCAN_CLK * 1000;
	candev->chip[chipnr]->chip_irq = TSCAN_IRQ ;
	candev->chip[chipnr]->chip_base_addr = 0x0;

	return 0;
}

int tscan_init_obj_data(struct canchip_t *chip, int objnr)
{
	chip->msgobj[objnr]->obj_base_addr = chip->chip_base_addr;
	return 0;
}

int tscan_program_irq(struct candevice_t *candev)
{
	return 0;
}

void tscan_write_register(unsigned data, can_ioptr_t address)
{
	can_spin_irqflags_t flags;
        can_spin_lock_irqsave(&tscan_win_lock,flags);
	//printk("write: 0x%x to 0x%x\n",data,address);
	*tscan_addr_io = (unsigned char)address;
	*tscan_data_io = (unsigned char)data;
	can_spin_unlock_irqrestore(&tscan_win_lock,flags);	
}

unsigned tscan_read_register(can_ioptr_t address)
{
	unsigned char data;
	can_spin_irqflags_t flags;
        can_spin_lock_irqsave(&tscan_win_lock,flags);
	*tscan_addr_io = (unsigned char)address;
	data = (unsigned char)*tscan_data_io;
	//printk("read: 0x%x from 0x%x\n",(unsigned char)*tscan_data_io,address);
	can_spin_unlock_irqrestore(&tscan_win_lock,flags);
	return data;
}

int tscan_register(struct hwspecops_t *hwspecops)
{
	hwspecops->request_io = tscan_request_io;
	hwspecops->release_io = tscan_release_io;
	hwspecops->reset = tscan_reset;
	hwspecops->init_hw_data = tscan_init_hw_data;
	hwspecops->init_chip_data = tscan_init_chip_data;
	hwspecops->init_obj_data = tscan_init_obj_data;
	hwspecops->write_register = tscan_write_register;
	hwspecops->read_register = tscan_read_register;
	hwspecops->program_irq = tscan_program_irq;
	return 0;
}

